"""
Copyright (C) 2013-2017 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

examples.py
~~~~~~~~~~~

Example models.

"""

import os

from . import core


PATHS = {
    'NationalScale': os.path.join(os.path.dirname(__file__), 'example_models', 'national_scale'),
    'UrbanScale': os.path.join(os.path.dirname(__file__), 'example_models', 'urban_scale'),
}


class NationalScale(core.Model):
    """
    National-scale example model.

    """

    def __init__(self, override=None):
        config_run = os.path.join(PATHS['NationalScale'], 'run.yaml')
        super().__init__(config_run=config_run, override=override)


class UrbanScale(core.Model):
    """
    Urban-scale example model.

    """

    def __init__(self, override=None):
        config_run = os.path.join(PATHS['UrbanScale'], 'run.yaml')
        super().__init__(config_run=config_run, override=override)



import tempfile

from calliope.utils import AttrDict
from . import common
from .common import assert_almost_equal, solver, solver_io


def create_and_run_model(override=""):
    locations = """
        locations:
            1:
                techs: ['ccgt', 'test_conversion', 'test_conversion_plus',
                        'demand_power', 'unmet_demand_power', 'demand_heat',
                        'demand_low_T', 'demand_V_low_T', 'supply_gas',
                        'unmet_demand_heat', 'unmet_demand_low_T',
                        'unmet_demand_V_low_T']
                override:
                    ccgt:
                        constraints:
                            e_cap.max: 30
                    test_conversion:
                        constraints:
                            e_cap.max: 20
                    demand_power:
                        constraints:
                            r: -10
                    demand_heat:
                        constraints:
                            r: -6
            2:
                techs: ['demand_V_low_T', 'unmet_demand_V_low_T']
                override:
                    demand_low_T:
                        constraints:
                            e_cap.max: 0
    """
    config_run = """
        mode: plan
        model: ['{techs}', '{locations}']
        subset_t: ['2005-01-01', '2005-01-01']
    """

    override = AttrDict.from_yaml_string(override)
    override.set_key('solver', solver)
    override.set_key('solver_io', solver_io)
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(locations.encode('utf-8'))
        f.read()
        model = common.simple_model(config_run=config_run,
                                    config_locations=f.name,
                                    override=override)
    model.run()
    return model


class TestModel:
    def test_model_conversion(self):
        model = create_and_run_model()
        assert str(model.results.solver.termination_condition) == 'optimal'

    def test_model_conversion_plus(self):
        override = """
            override.locations.1.override.demand_low_T.constraints.r: -5
            override.locations.1.override.demand_V_low_T.constraints.r: -5
        """
        model = create_and_run_model(override)
        assert str(model.results.solver.termination_condition) == 'optimal'
        sol = model.solution
        assert_almost_equal(sol.costs.sum(), 41.47, tolerance=0.1)
        assert_almost_equal(sol.c_prod.loc[
            dict(y='test_conversion_plus', c='power', x='1')], 13.5, tolerance=0.01)
        assert_almost_equal(sol.c_prod.loc[
            dict(y='test_conversion_plus', c='power', x='2')], 0, tolerance=0.01)

    def test_model_conversion_plus_leakage(self):
        """
        Check that non-primary carriers of conversion_plus techs aren't
        erroneously producing in locations where they're not allowed
        """
        override = """
            override.locations.2.override.demand_V_low_T.constraints.r: -5
        """
        model = create_and_run_model(override)
        assert str(model.results.solver.termination_condition) == 'optimal'
        sol = model.solution
        assert_almost_equal(sol.c_prod.loc[
            dict(y='test_conversion_plus', c='V_low_T', x='2')], 0, tolerance=0.01)

    def test_model_conversion_plus_e_cap(self):
        """
        Check that non-primary carriers of conversion_plus techs, defined in
        `carrier_out` aren't erroneously producing above the set e_cap of the
        technology
        """
        override = """
            override.locations.1.override.demand_low_T.constraints.r: 0
            override.locations.1.override.demand_V_low_T.constraints:
                                    r: -5
                                    e_cap.max: 20
            override.locations.1.override.test_conversion_plus.constraints:
                                    e_cap.max: 0
        """
        model = create_and_run_model(override)
        assert str(model.results.solver.termination_condition) == 'optimal'
        sol = model.solution
        assert_almost_equal(sol.c_prod.loc[
            dict(y='test_conversion_plus', c='low_T', x='1')], 0)
        assert_almost_equal(sol.c_prod.loc[
            dict(y='test_conversion_plus', c='V_low_T', x='1')], 0)



"""
Copyright (C) 2013-2017 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

objective.py
~~~~~~~~~~~~

Objective functions.

"""

import pyomo.core as po  # pylint: disable=import-error

def objective_cost_minimization(model):
    """
    Minimizes total system monetary cost.
    Used as a default if a model does not specify another objective.

    """
    m = model.m

    def obj_rule(m):
        return sum(model.get_option(y + '.weight') *
                       sum(m.cost[y, x, 'monetary']
                        for x in m.x) for y in m.y)

    m.obj = po.Objective(sense=po.minimize, rule=obj_rule)
    m.obj.domain = po.Reals



import tempfile

from calliope.utils import AttrDict
from . import common
from .common import assert_almost_equal, solver, solver_io


def create_and_run_model(override=""):
    locations = """
        locations:
            1:
                techs: ['ccgt','demand_power', 'unmet_demand_power']
                override:
                    ccgt:
                        constraints:
                            e_cap.max: 100
                    demand_power:
                        x_map: '1: demand'
                        constraints:
                            r: file=demand-sin_r.csv
        links:
    """
    config_run = """
        mode: plan
        model: ['{techs}', '{locations}']
        subset_t: ['2005-01-01', '2005-01-03']
    """

    override = AttrDict.from_yaml_string(override)
    override.set_key('solver', solver)
    override.set_key('solver_io', solver_io)
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(locations.encode('utf-8'))
        f.read()
        model = common.simple_model(config_run=config_run,
                                    config_locations=f.name,
                                    override=override)
    model.run()
    return model


class TestModel:
    def test_system_margin(self):
        override1 = """
            override.system_margin.power: 0
        """
        override2 = """
            override.system_margin.power: 0.5
        """
        model1 = create_and_run_model(override1)
        model2 = create_and_run_model(override2)
        assert str(model1.results.solver.termination_condition) == 'optimal'
        assert str(model2.results.solver.termination_condition) == 'optimal'
        sol1 = model1.solution
        assert_almost_equal(sol1['e'].loc[dict(c='power', y='ccgt')].sum(), 548.79, tolerance=0.01)
        # assert_almost_equal(sol1['e_cap'].loc[dict(y='ccgt')].sum(), 9.5, tolerance=0.01)
        assert_almost_equal(sol1['e_cap'].sum(), 20.00, tolerance=0.01)
        sol2 = model2.solution
        assert_almost_equal(sol2['e'].loc[dict(c='power', y='ccgt')].sum(), 548.79, tolerance=0.01)
        assert_almost_equal(sol2['e_cap'].sum(), 25.00, tolerance=0.01)



import tempfile

import pytest

from calliope.utils import AttrDict
from calliope.exceptions import ModelError
from . import common
from .common import assert_almost_equal, solver, solver_io


def create_and_run_model(override=""):
    locations = """
        locations:
            1:
                techs: ['ccgt', 'test_storage', 'demand_power',
                        'unmet_demand_power']
                override:
                    ccgt:
                        constraints:
                            e_cap.max: 9.5
                    demand_power:
                        x_map: '1: demand'
                        constraints:
                            r: file=demand-sin_r.csv
        links:
    """
    config_run = """
        mode: plan
        model: ['{techs}', '{locations}']
        subset_t: ['2005-01-01', '2005-01-03']
    """

    override = AttrDict.from_yaml_string(override)
    override.set_key('solver', solver)
    override.set_key('solver_io', solver_io)
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(locations.encode('utf-8'))
        f.read()
        model = common.simple_model(config_run=config_run,
                                    config_locations=f.name,
                                    override=override)
    model.run()
    return model


class TestModel:
    def test_model_e_cap_max(self):
        override = """
            override.techs.test_storage.constraints.e_cap.max: 0.6
        """
        model = create_and_run_model(override)
        assert str(model.results.solver.termination_condition) == 'optimal'
        sol = model.solution
        assert_almost_equal(sol['e'].loc[dict(c='power', y='ccgt')].sum(dim='x')[dict(t=0)].mean(),
                            8.03, tolerance=0.01)
        assert_almost_equal(sol['e'].loc[dict(c='power', y='ccgt')].sum(dim='x').mean(),
                            7.62, tolerance=0.01)

    def test_model_c_rate(self):
        override = """
            override.techs.test_storage.constraints.c_rate: 0.006
        """
        model = create_and_run_model(override)
        assert str(model.results.solver.termination_condition) == 'optimal'
        sol = model.solution
        assert_almost_equal(sol['e'].loc[dict(c='power', y='ccgt')].sum(dim='x')[dict(t=0)].mean(),
                            8.03, tolerance=0.01)
        assert_almost_equal(sol['e'].loc[dict(c='power', y='ccgt')].sum(dim='x').mean(),
                            7.62, tolerance=0.01)

    def test_model_compare_c_rate_e_cap(self):
        override1 = """
            override.techs.test_storage.constraints:
                            c_rate: 0.006
                            e_cap.max: 0.6
        """
        override2 = """
            override.techs.test_storage.constraints:
                            c_rate: 0.008
                            e_cap.max: 0.6
        """
        model1 = create_and_run_model(override1)
        model2 = create_and_run_model(override2)
        sol1 = model1.solution
        sol2 = model2.solution
        assert_almost_equal(sol1.e_cap.loc[dict(y='test_storage')].sum(dim='x'),
                            sol2.e_cap.loc[dict(y='test_storage')].sum(dim='x'),
                            tolerance=0.01)
        assert (sol1.s_cap.loc[dict(y='test_storage')].sum(dim='x') >
                sol2.s_cap.loc[dict(y='test_storage')].sum(dim='x'))

    def test_model_s_time(self):
        override1 = """
            override.techs.test_storage.constraints:
                            s_time.max: 12
                            c_rate: 0.01
        """
        override2 = """
            override.techs.test_storage.constraints:
                            s_time.max: 12
                            e_cap.max: 1
        """
        model1 = create_and_run_model(override1)
        model2 = create_and_run_model(override2)
        assert str(model1.results.solver.termination_condition) == 'optimal'
        assert str(model2.results.solver.termination_condition) == 'optimal'
        sol1 = model1.solution
        sol2 = model2.solution
        assert (sol1.e_cap.loc[dict(y='test_storage')].sum(dim='x') <
                sol2.e_cap.loc[dict(y='test_storage')].sum(dim='x'))
        assert (sol1.s_cap.loc[dict(y='test_storage')].sum(dim='x') >
                sol2.s_cap.loc[dict(y='test_storage')].sum(dim='x'))

    def test_model_c_rate_and_e_cap_and_s_cap_dontmatch(self):
        override = """
            override.techs.test_storage.constraints:
                            c_rate: 0.006
                            e_cap.equals: 0.6
                            s_cap.equals: 1000
        """
        with pytest.raises(ModelError):
            model = create_and_run_model(override)

    def test_model_c_rate_and_e_cap_and_s_cap_match(self):
        override = """
            override.techs.test_storage.constraints:
                            c_rate: 0.1
                            e_cap.equals: 10
                            s_cap.equals: 100
        """
        # Just make sure this raises no error
        model = create_and_run_model(override)

    def test_model_e_cap_and_s_cap_only(self):
        override = """
            override.techs.test_storage.constraints:
                            e_cap.max: 0.6
                            s_cap.max: 100
        """
        # Just make sure this raises no error
        model = create_and_run_model(override)



"""
Copyright (C) 2013-2017 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

base_planning.py
~~~~~~~

Basic model constraints.

"""

import pyomo.core as po  # pylint: disable=import-error
import numpy as np
import xarray as xr

from .. import exceptions
from .. import transmission
from .. import utils


def get_constraint_param(model, param_string, y, x, t):
    """
    Function to get values for constraints which can optionally be
    loaded from file (so may have time dependency).

    model = calliope model
    param_string = constraint as string
    y = technology
    x = location
    t = timestep
    """

    if param_string in model.data and y in model._sets['y_' + param_string + '_timeseries']:
        return getattr(model.m, param_string + '_param')[y, x, t]
    else:
        return model.get_option(y + '.constraints.' + param_string, x=x)


def get_cost_param(model, param_string, k, y, x, t):
    """
    Function to get values for constraints which can optionally be
    loaded from file (so may have time dependency).

    model = calliope model
    cost = cost name, e.g. 'om_fuel'
    k = cost type, e.g. 'monetary'
    y = technology
    x = location
    t = timestep
    """

    cost_getter = utils.cost_getter(model.get_option)

    @utils.memoize
    def _cost(cost, y, k, x=None):
        return cost_getter(cost, y, k, x=x)

    if param_string in model.data and y in model._sets['y_' + param_string + '_timeseries']:
        return getattr(model.m, param_string + '_param')[y, x, t, k]
    else:  # Search in model.config_model
        return _cost(param_string, y, k, x=x)


def generate_variables(model):
    """
    Defines variables:

    * r: resource -> tech (+ production
    * r_area: resource collector area
    * r2: secondary resource -> storage (+ production)
    * c_prod: tech -> carrier (+ production)
    * c_con: tech <- carrier (- consumption)
    * s_cap: installed storage capacity
    * r_cap: installed resource <-> storage conversion capacity
    * e_cap: installed storage <-> grid conversion capacity (gross)
    * r2_cap: installed secondary resource conversion capacity

    * cost: total costs
    * cost_con: construction costs
    * cost_op_fixed: fixed operation costs
    * cost_op_var: variable operation costs
    * cost_op_fuel: primary resource fuel costs
    * cost_op_r2: secondary resource fuel costs

    """

    m = model.m

    # Capacity
    m.r_area = po.Var(m.y_r_area, m.x_r, within=po.NonNegativeReals)
    m.s_cap = po.Var(m.y_store, m.x_store, within=po.NonNegativeReals)
    m.r_cap = po.Var(m.y_sp_finite_r, m.x_r, within=po.NonNegativeReals)  # FIXME maybe should be y_finite_r?
    m.e_cap = po.Var(m.y, m.x, within=po.NonNegativeReals)
    m.r2_cap = po.Var(m.y_sp_r2, m.x_r, within=po.NonNegativeReals)

    # Unit commitment
    m.r = po.Var(m.y_sp_finite_r, m.x_r, m.t, within=po.Reals)
    m.r2 = po.Var(m.y_sp_r2, m.x_r, m.t, within=po.NonNegativeReals)
    m.s = po.Var(m.y_store, m.x_store, m.t, within=po.NonNegativeReals)
    m.c_prod = po.Var(m.c, m.y, m.x, m.t, within=po.NonNegativeReals)
    m.c_con = po.Var(m.c, m.y, m.x, m.t, within=po.NegativeReals)
    m.export = po.Var(m.y_export, m.x_export, m.t, within=po.NonNegativeReals)

    # Costs
    m.cost_var = po.Var(m.y, m.x, m.t, m.k, within=po.Reals)
    m.cost_fixed = po.Var(m.y, m.x, m.k, within=po.Reals)
    m.cost = po.Var(m.y, m.x, m.k, within=po.Reals)


def node_resource(model):
    m = model.m

    # TODO reformulate c_r_rule conditionally once Pyomo supports that.
    # had to remove the following formulation because it is not
    # re-evaluated on model re-construction -- we now check for
    # demand/supply tech instead, which means that `r` can only
    # be ALL negative or ALL positive for a given tech!
    # Ideally we have `elif po.value(m.r[y, x, t]) > 0:` instead of
    # `elif y in m.y_supply or y in m.y_unmet_demand:` and `elif y in m.y_demand:`

    def r_available_rule(m, y, x, t):
        r_scale = model.get_option(y + '.constraints.r_scale', x=x)
        force_r = get_constraint_param(model, 'force_r', y, x, t)

        if y in m.y_sd:
            e_eff = get_constraint_param(model, 'e_eff', y, x, t)
            if po.value(e_eff) == 0:
                c_prod = 0
            else:
                c_prod = sum(m.c_prod[c, y, x, t] for c in m.c) / e_eff # supply techs
            c_con = sum(m.c_con[c, y, x, t] for c in m.c) * e_eff # demand techs

            if y in m.y_sd_r_area:
                r_avail = (get_constraint_param(model, 'r', y, x, t) * r_scale
                           * m.r_area[y, x])
            else:
                r_avail = (get_constraint_param(model, 'r', y, x, t) * r_scale)

            if force_r:
                return c_prod + c_con == r_avail
            else:
                return c_prod - c_con <= r_avail

        elif y in m.y_supply_plus:
            r_eff = get_constraint_param(model, 'r_eff', y, x, t)

            if y in m.y_sp_r_area:
                r_avail = (get_constraint_param(model, 'r', y, x, t) * r_scale
                           * m.r_area[y, x] * r_eff)
            else:
                r_avail = get_constraint_param(model, 'r', y, x, t) * r_scale * r_eff

            if force_r:
                return m.r[y, x, t] == r_avail
            else:
                return m.r[y, x, t] <= r_avail

    # Constraints
    m.c_r_available = po.Constraint(m.y_finite_r, m.x_r, m.t,
                                    rule=r_available_rule)


def node_energy_balance(model):
    m = model.m
    time_res = model.data['_time_res'].to_series()

    def get_e_eff_per_distance(model, y, x):
        e_loss = model.get_option(y + '.constraints_per_distance.e_loss', x=x)
        per_distance = model.get_option(y + '.per_distance')
        tech, x2 = y.split(':')
        link = model.config_model.get_key(
            'links.' + x + ',' + x2,
            default=model.config_model['links'].get(x2 + ',' + x)
        )
        # link = None if no link exists
        if not link or tech not in link.keys():
            return 1.0
        try:
            distance = link.get_key(tech + '.distance')
        except KeyError:
            if e_loss > 0:
                e = exceptions.OptionNotSetError
                raise e('Distance must be defined for link: {} '
                        'and transmission tech: {}, as e_loss per distance '
                        'is defined'.format(x + ',' + x2, tech))
            else:
                return 1.0
        return 1 - (e_loss * (distance / per_distance))

    def get_conversion_out(c_1, c_2, m, y, x, t):
        if isinstance(c_1, dict):
            c_prod1 = sum([m.c_prod[c, y, x, t] / c_1[c] for c in c_1.keys()])
        else:
            c_prod1 = m.c_prod[c_1, y, x, t]
        if isinstance(c_2, dict):
            c_min = min(c_2.values())
            c_prod2 = sum([m.c_prod[c, y, x, t] / (c_2[c] / c_min) for c in c_2.keys()])
        else:
            c_min = 1
            c_prod2 = m.c_prod[c_2, y, x, t]
        return c_prod1 * c_min == c_prod2

    def get_conversion_in(c_1, c_2, m, y, x, t):
        if isinstance(c_1, dict):
            c_con1 = sum([m.c_con[c, y, x, t] / c_1[c] for c in c_1.keys()])
        else:
            c_con1 = m.c_con[c_1, y, x, t]
        if isinstance(c_2, dict):
            c_min = min(c_2.values())
            c_con2 = sum([m.c_con[c, y, x, t] / (c_2[c] / c_min) for c in c_2.keys()])
        else:
            c_min = 1
            c_con2 = m.c_con[c_2, y, x, t]
        return c_con1 * c_min == c_con2

    # Constraint rules
    def transmission_rule(m, y, x, t):
        y_remote, x_remote = transmission.get_remotes(y, x)
        e_eff = get_constraint_param(model, 'e_eff', y, x, t)
        if y_remote in m.y_transmission:
            c = model.get_option(y + '.carrier')

            return (m.c_prod[c, y, x, t]
                    == -1 * m.c_con[c, y_remote, x_remote, t]
                    * e_eff
                    * get_e_eff_per_distance(model, y, x))
        else:
            return po.Constraint.NoConstraint

    def conversion_rule(m, y, x, t):
        c_out = model.get_option(y + '.carrier_out')
        c_in = model.get_option(y + '.carrier_in')
        e_eff = get_constraint_param(model, 'e_eff', y, x, t)
        return (m.c_prod[c_out, y, x, t]
                == -1 * m.c_con[c_in, y, x, t] * e_eff)

    def conversion_plus_primary_rule(m, y, x, t):
        c_out = model.get_option(y + '.carrier_out')
        c_in = model.get_option(y + '.carrier_in')
        e_eff = get_constraint_param(model, 'e_eff', y, x, t)
        if isinstance(c_out, dict):
            c_prod = sum(m.c_prod[c, y, x, t] / c_out[c] for c in c_out.keys())
        else:
            c_prod = m.c_prod[c_out, y, x, t]
        if isinstance(c_in, dict):
            c_con = sum([m.c_con[c, y, x, t] * c_in[c] for c in c_in.keys()])
        else:
            c_con = m.c_con[c_in, y, x, t]
        return c_prod == -1 * c_con * e_eff

    def conversion_plus_secondary_out_rule(m, y, x, t):
        c_1 = model.get_option(y + '.carrier_out')
        c_2 = model.get_option(y + '.carrier_out_2')
        return get_conversion_out(c_1, c_2, m, y, x, t)

    def conversion_plus_tertiary_out_rule(m, y, x, t):
        c_1 = model.get_option(y + '.carrier_out')
        c_3 = model.get_option(y + '.carrier_out_3')
        return get_conversion_out(c_1, c_3, m, y, x, t)

    def conversion_plus_secondary_in_rule(m, y, x, t):
        c_1 = model.get_option(y + '.carrier_in')
        c_2 = model.get_option(y + '.carrier_in_2')
        return get_conversion_in(c_1, c_2, m, y, x, t)

    def conversion_plus_tertiary_in_rule(m, y, x, t):
        c_1 = model.get_option(y + '.carrier_in')
        c_3 = model.get_option(y + '.carrier_in_3')
        return get_conversion_in(c_1, c_3, m, y, x, t)

    def supply_plus_rule(m, y, x, t):
        e_eff = get_constraint_param(model, 'e_eff', y, x, t)
        p_eff = model.get_option(y + '.constraints.p_eff', x=x)
        total_eff = e_eff * p_eff
        # TODO once Pyomo supports it,
        # let this update conditionally on param update!
        if po.value(total_eff) == 0:
            c_prod = 0
        else:
            c_prod = (sum(m.c_prod[c, y, x, t] for c in m.c)) / total_eff

        # If this tech is in the set of techs allowing r2, include it
        if y in m.y_sp_r2:
            r2 = m.r2[y, x, t]
        else:
            r2 = 0

        # A) Case where no storage allowed
        if y not in m.y_store or x not in m.x_store:
            return m.r[y, x, t] == c_prod - r2

        # B) Case where storage is allowed
        r = m.r[y, x, t]
        if m.t.order_dict[t] == 0:
            s_minus_one = m.s_init[y, x]
        else:
            s_loss = get_constraint_param(model, 's_loss', y, x, t)
            s_minus_one = (((1 - s_loss)
                            ** time_res.at[model.prev_t(t)])
                           * m.s[y, x, model.prev_t(t)])
        return (m.s[y, x, t] == s_minus_one + r + r2 - c_prod)

    def storage_rule(m, y, x, t):
        e_eff = get_constraint_param(model, 'e_eff', y, x, t)
        p_eff = model.get_option(y + '.constraints.p_eff', x=x)
        total_eff = e_eff * p_eff

        # TODO once Pyomo supports it,
        # let this update conditionally on param update!
        if po.value(total_eff) == 0:
            c_prod = 0
        else:
            c_prod = sum(m.c_prod[c, y, x, t] for c in m.c) / total_eff
        c_con = sum(m.c_con[c, y, x, t] for c in m.c) * total_eff

        if m.t.order_dict[t] == 0:
            s_minus_one = m.s_init[y, x]
        else:
            s_loss = get_constraint_param(model, 's_loss', y, x, t)
            s_minus_one = (((1 - s_loss)
                            ** time_res.at[model.prev_t(t)])
                           * m.s[y, x, model.prev_t(t)])
        return (m.s[y, x, t] == s_minus_one - c_prod - c_con)

    # Constraints
    m.c_balance_transmission = po.Constraint(m.y_transmission, m.x_transmission, m.t,
                                            rule=transmission_rule)
    m.c_balance_conversion = po.Constraint(m.y_conversion, m.x_conversion, m.t,
                                            rule=conversion_rule)
    m.c_balance_conversion_plus = po.Constraint(m.y_conversion_plus, m.x_conversion, m.t,
                                            rule=conversion_plus_primary_rule)
    m.c_balance_conversion_plus_secondary_out = po.Constraint(m.y_cp_2out, m.x_conversion, m.t,
                                            rule=conversion_plus_secondary_out_rule)
    m.c_balance_conversion_plus_tertiary_out = po.Constraint(m.y_cp_3out, m.x_conversion, m.t,
                                            rule=conversion_plus_tertiary_out_rule)
    m.c_balance_conversion_plus_secondary_in = po.Constraint(m.y_cp_2in, m.x_conversion, m.t,
                                            rule=conversion_plus_secondary_in_rule)
    m.c_balance_conversion_plus_tertiary_in = po.Constraint(m.y_cp_3in, m.x_conversion, m.t,
                                            rule=conversion_plus_tertiary_in_rule)
    m.c_balance_supply_plus = po.Constraint(m.y_supply_plus, m.x_r, m.t,
                                            rule=supply_plus_rule)
    m.c_balance_storage = po.Constraint(m.y_storage, m.x_store, m.t,
                                            rule=storage_rule)


def node_constraints_build(model):

    m = model.m

    def get_var_constraint(model_var, y, var, x,
                           _equals=None, _max=None, _min=None,
                           scale=None):

        if not _equals:
            _equals = model.get_option(y + '.constraints.'
                                       + var + '.equals', x=x)
        if not _max:
            _max = model.get_option(y + '.constraints.' + var + '.max', x=x)
        if not _min:
            _min = model.get_option(y + '.constraints.' + var + '.min', x=x)
        if scale:
            _equals = scale * _equals
            _min = scale * _min
            _max = scale * _max
        if _equals:
            if np.isinf(_equals):
                e = exceptions.ModelError
                raise e('Cannot use inf in operational mode, for value of '
                        '{}.{}.equals.{}'.format(y, var, x))
            return model_var == _equals
        elif model.mode == 'operate':
            # Operational mode but 'equals' constraint not set, we use 'max'
            # instead
            # FIXME this should be logged
            if np.isinf(_max):
                return po.Constraint.NoConstraint
            else:
                return model_var == _max
        else:
            if np.isinf(_max):
                _max = None  # to disable upper bound
            if _min == 0 and _max is None:
                return po.Constraint.NoConstraint
            else:
                return (_min, model_var, _max)

    def get_s_cap(y, x, e_cap, c_rate):
        """
        Get s_cap.max from s_time.max, where applicable. If s_time.max is used,
        the maximum storage possible is the minimum value of storage possible for
        any time length which meets the s_time.max value.
        """
        # TODO:
        # incorporate timeseries resolution. Currently assumes that each
        # timestep is worth one unit of time.
        s_time_max = model.get_option(y + '.constraints.s_time.max', x=x)
        s_cap_max = model.get_option(y + '.constraints.s_cap.max', x=x)
        if not s_cap_max:
            s_cap_max = np.inf
        if not s_time_max:
            return s_cap_max
        if not e_cap and not c_rate:
            return 0
        elif not e_cap:
            e_cap = s_cap_max * c_rate
        elif e_cap and c_rate:
            e_cap = min(e_cap, s_cap_max * c_rate)

        s_loss = model.data.loc[dict(y=y, x=x)].get('s_loss',
                                model.get_option(y+ '.constraints.s_loss', x=x))
        e_eff = model.data.loc[dict(y=y, x=x)].get('e_eff',
                                model.get_option(y+ '.constraints.e_eff', x=x))
        try:
            leakage = 1 / (1 - s_loss)  # if s_loss is timeseries dependant, will be a DataArray
            discharge = e_cap / e_eff  # if e_eff is timeseries dependant, will be a DataArray
        except ZeroDivisionError:  # if s_loss = 1 or e_eff = 0
            return np.inf  # i.e. no upper limit on storage
        exponents = [i for i in range(s_time_max)]
        if isinstance(leakage, xr.DataArray) or isinstance(discharge, xr.DataArray):
            roll = model.data.t.rolling(t=s_time_max)  # create arrays of all rolling horizons
            S = {s_cap_max}
            for label, arr_window in roll:
                if len(arr_window) == s_time_max:  # only consider arrays of the maximum time length
                    S.add(sum(discharge * np.power(leakage, exponents)))
            return min(S)  # smallest value of storage is the maximum allowed
        else:  # no need to loop through rolling horizons if all values are static in time
            return min(s_cap_max, sum(discharge * np.power(leakage, exponents)))

    # Constraint rules
    def c_s_cap_rule(m, y, x):
        """
        Set maximum storage capacity. Supply_plus & storage techs only
        This can be set by either s_cap (kWh) or by
        e_cap (charge/discharge capacity) * charge rate.
        If s_cap.equals and e_cap.equals are set for the technology, then
        s_cap * charge rate = e_cap must hold. Otherwise, take the lowest capacity
        capacity defined by s_cap.max or e_cap.max / charge rate.
        """
        s_cap_equals = model.get_option(y + '.constraints.s_cap.equals', x=x)
        scale = model.get_option(y + '.constraints.e_cap_scale', x=x)
        e_cap = model.get_option(y + '.constraints.e_cap.equals', x=x) * scale
        charge_rate = model.get_option(y + '.constraints.c_rate', x=x)
        if e_cap and s_cap_equals and charge_rate and s_cap_equals * charge_rate != e_cap:
            raise exceptions.ModelError(
                'e_cap.equals and s_cap.equals must '
                'be equivalent when considering charge rate for {}:{}'
                .format(y, x)
            )
        if not e_cap:
            e_cap = model.get_option(y + '.constraints.e_cap.max', x=x) * scale
        if not s_cap_equals:
            s_cap_max = get_s_cap(y, x, e_cap, charge_rate)
            return get_var_constraint(m.s_cap[y, x], y, 's_cap', x, _max=s_cap_max)
        else:
            return get_var_constraint(m.s_cap[y, x], y, 's_cap', x, _equals=s_cap_equals)

    def c_r_cap_rule(m, y, x):
        if model.get_option(y + '.constraints.r_cap_equals_e_cap', x=x):
            return m.r_cap[y, x] == m.e_cap[y, x]
        else:
            return get_var_constraint(m.r_cap[y, x], y, 'r_cap', x)

    def c_r_area_rule(m, y, x):
        """
        Set maximum r_area. Supply_plus techs only.
        """
        area_per_cap = model.get_option(y + '.constraints.r_area_per_e_cap', x=x)
        if area_per_cap:
            return m.r_area[y, x] == m.e_cap[y, x] * area_per_cap
        else:
            e_cap_max = model.get_option(y + '.constraints.e_cap.max', x=x)
            if e_cap_max == 0:
                # If a technology has no e_cap here, we force r_area to zero,
                # so as not to accrue spurious costs
                return m.r_area[y, x] == 0
            else:
                return get_var_constraint(m.r_area[y, x], y, 'r_area', x)

    def c_e_cap_rule(m, y, x):
        """
        Set maximum e_cap. All technologies.
        """
        # First check whether this tech is allowed at this location
        if not model._locations.at[x, y] == 1:
            return m.e_cap[y, x] == 0
        e_cap_scale = model.get_option(y + '.constraints.e_cap_scale', x=x)
        if y in m.y_transmission:
            e_cap_equals = model._locations.get(
                '_override.{}.constraints.e_cap.equals'.format(y), {x: None})[x]
            e_cap_max = model._locations.get(
                '_override.{}.constraints.e_cap.max'.format(y), {x: None})[x]
        else:
            e_cap_max = None
            e_cap_equals = None
        return get_var_constraint(m.e_cap[y, x], y, 'e_cap', x, scale=e_cap_scale,
                                  _max=e_cap_max, _equals=e_cap_equals)

    def c_e_cap_storage_rule(m, y, x):
        """
        Set maximum e_cap. All storage technologies.
        """
        e_cap_scale = model.get_option(y + '.constraints.e_cap_scale', x=x)
        charge_rate = model.get_option(y + '.constraints.c_rate', x=x)
        if charge_rate:
            e_cap_max = m.s_cap[y, x] * charge_rate * e_cap_scale
            return m.e_cap[y, x] <= e_cap_max
        else:
            return po.Constraint.Skip

    def c_r2_cap_rule(m, y, x):
        """
        Set secondary resource capacity. Supply_plus techs only.
        """
        follow = model.get_option(y + '.constraints.r2_cap_follow', x=x)
        mode = model.get_option(y + '.constraints.r2_cap_follow_mode', x=x)

        # First deal with the special case of ``r2_cap_follow`` being set
        if follow:
            try:
                r2_cap_val = getattr(m, follow)[y, x]
            except AttributeError:
                # Raise an error to make sure follows isn't accidentally set to
                # something invalid
                e = exceptions.ModelError
                raise e('r2_cap_follow set to invalid value at '
                        '({}, {}): {}'.format(y, x, follow))

            if mode == 'max':
                return m.r2_cap[y, x] <= r2_cap_val
            elif mode == 'equals':
                return m.r2_cap[y, x] == r2_cap_val

        else:  # If ``r2_cap_follow`` not set, set up standard constraints
            return get_var_constraint(m.r2_cap[y, x], y, 'r2_cap', x)

    # Constraints
    m.c_s_cap = po.Constraint(m.y_store, m.x_store, rule=c_s_cap_rule)
    m.c_r_cap = po.Constraint(m.y_sp_finite_r, m.x_r, rule=c_r_cap_rule)
    m.c_r_area = po.Constraint(m.y_r_area, m.x_r, rule=c_r_area_rule)
    m.c_e_cap = po.Constraint(m.y, m.x, rule=c_e_cap_rule)
    m.c_e_cap_storage = po.Constraint(m.y_store, m.x_store, rule=c_e_cap_storage_rule)
    m.c_r2_cap = po.Constraint(m.y_sp_r2, m.x_r, rule=c_r2_cap_rule)


def node_constraints_operational(model):
    m = model.m
    time_res = model.data['_time_res'].to_series()

    # Constraint rules
    def r_max_upper_rule(m, y, x, t):
        """
        set maximum resource supply. Supply_plus techs only.
        """
        return m.r[y, x, t] <= time_res.at[t] * m.r_cap[y, x]

    def c_prod_max_rule(m, c, y, x, t):
        """
        Set maximum carrier production. All technologies.
        """
        allow_c_prod = get_constraint_param(model, 'e_prod', y, x, t)
        c_prod = m.c_prod[c, y, x, t]
        if y in m.y_conversion_plus:  # Conversion techs with 2 or more output carriers
            carriers_out = model.get_carrier(y, 'out', all_carriers=True)
            if c not in carriers_out:
                return c_prod == 0
            if c in carriers_out and model._locations.at[x, y] == 0:
                return c_prod == 0
            else:
                return po.Constraint.Skip
        if not allow_c_prod:
            return c_prod == 0
        p_eff = model.get_option(y + '.constraints.p_eff', x=x)
        if c == model.get_option(y + '.carrier', default=y + '.carrier_out'):
            return c_prod <= time_res.at[t] * m.e_cap[y, x] * p_eff
        else:
            return m.c_prod[c, y, x, t] == 0

    def c_prod_min_rule(m, c, y, x, t):
        """
        Set minimum carrier production. All technologies except conversion_plus
        """
        min_use = get_constraint_param(model, 'e_cap_min_use', y, x, t)
        allow_c_prod = get_constraint_param(model, 'e_prod', y, x, t)
        if not min_use or not allow_c_prod:
            return po.Constraint.NoConstraint
        if y in m.y_conversion_plus:  # Conversion techs with 2 output carriers
            return po.Constraint.Skip
        elif c == model.get_option(y + '.carrier', default=y + '.carrier_out'):
            return (m.c_prod[c, y, x, t]
                    >= time_res.at[t] * m.e_cap[y, x] * min_use)
        else:
            return po.Constraint.Skip

    def c_conversion_plus_prod_max_rule(m, y, x, t):
        """
        Set maximum carrier production. Conversion_plus technologies.
        """
        allow_c_prod = get_constraint_param(model, 'e_prod', y, x, t)
        c_out = model.get_option(y + '.carrier_out')
        e_eff = get_constraint_param(model, 'e_eff', y, x, t)
        if isinstance(c_out, dict):
            c_prod = sum(m.c_prod[c, y, x, t] for c in c_out.keys())
        else:
            c_prod = m.c_prod[c_out, y, x, t]
        if not allow_c_prod:
            return c_prod == 0
        else:
            return c_prod <= time_res.at[t] * m.e_cap[y, x]

    def c_conversion_plus_prod_min_rule(m, y, x, t):
        """
        Set minimum carrier production. Conversion_plus technologies.
        """
        min_use = get_constraint_param(model, 'e_cap_min_use', y, x, t)
        allow_c_prod = get_constraint_param(model, 'e_prod', y, x, t)
        c_out = model.get_option(y + '.carrier_out')
        if not min_use or not allow_c_prod:
            return po.Constraint.NoConstraint
        else:
            c_prod_min = time_res.at[t] * m.e_cap[y, x] * min_use
            if isinstance(c_out, dict):
                c_prod = sum(m.c_prod[c, y, x, t] for c in c_out.keys())
            else:
                c_prod = m.c_prod[c_out, y, x, t]
            return c_prod >= c_prod_min

    def c_con_max_rule(m, c, y, x, t):
        """
        Set maximum carrier consumption. All technologies.
        """
        allow_c_con = get_constraint_param(model, 'e_con', y, x, t)
        p_eff = model.get_option(y + '.constraints.p_eff', x=x)
        if y in m.y_conversion or y in m.y_conversion_plus:
            carriers = model.get_cp_carriers(y, x, direction='in')[1]
            if c not in carriers:
                return m.c_con[c, y, x, t] == 0
            else:
                return po.Constraint.Skip
        if (allow_c_con is True and
                c == model.get_option(y + '.carrier', default=y + '.carrier_in')):
            return m.c_con[c, y, x, t] >= (-1 * time_res.at[t]
                                            * m.e_cap[y, x] * p_eff)
        else:
            return m.c_con[c, y, x, t] == 0

    def s_max_rule(m, y, x, t):
        """
        Set maximum stored energy. Supply_plus & storage techs only.
        """
        return m.s[y, x, t] <= m.s_cap[y, x]

    def r2_max_rule(m, y, x, t):
        """
        Set maximum secondary resource supply. Supply_plus techs only.
        """
        r2_startup = get_constraint_param(model, 'r2_startup_only', y, x, t)
        if (r2_startup and t >= model.data.startup_time_bounds):
            return m.r2[y, x, t] == 0
        else:
            return m.r2[y, x, t] <= time_res.at[t] * m.r2_cap[y, x]

    def c_export_max_rule(m, y, x, t):
        """
        Set maximum export. All exporting technologies.
        """
        if get_constraint_param(model, 'export_cap', y, x, t):
            return (m.export[y, x, t] <=
                    get_constraint_param(model, 'export_cap', y, x, t))

        else:
            return po.Constraint.Skip

    # Constraints
    m.c_r_max_upper = po.Constraint(
        m.y_sp_finite_r, m.x_r, m.t, rule=r_max_upper_rule)
    m.c_prod_max = po.Constraint(
        m.c, m.y, m.x, m.t, rule=c_prod_max_rule)
    m.c_prod_min = po.Constraint(
        m.c, m.y, m.x, m.t, rule=c_prod_min_rule)
    m.c_conversion_plus_prod_max = po.Constraint(
        m.y_conversion_plus, m.x, m.t, rule=c_conversion_plus_prod_max_rule)
    m.c_conversion_plus_prod_min = po.Constraint(
        m.y_conversion_plus, m.x, m.t, rule=c_conversion_plus_prod_min_rule)
    m.c_con_max = po.Constraint(
        m.c, m.y, m.x, m.t, rule=c_con_max_rule)
    m.c_s_max = po.Constraint(
        m.y_store, m.x_store, m.t, rule=s_max_rule)
    m.c_r2_max = po.Constraint(
        m.y_sp_r2, m.x_r, m.t, rule=r2_max_rule)
    m.c_export_max = po.Constraint(
        m.y_export, m.x_export, m.t, rule=c_export_max_rule)


def node_constraints_transmission(model):
    """
    Constrain e_cap symmetrically for transmission nodes. Transmission techs only.
    """
    m = model.m

    # Constraint rules
    def c_trans_rule(m, y, x):
        y_remote, x_remote = transmission.get_remotes(y, x)
        if y_remote in m.y_transmission:
            return m.e_cap[y, x] == m.e_cap[y_remote, x_remote]
        else:
            return po.Constraint.NoConstraint

    # Constraints
    m.c_transmission_capacity = po.Constraint(m.y_transmission, m.x_transmission,
                                              rule=c_trans_rule)


def node_costs(model):

    m = model.m
    time_res = model.data['_time_res'].to_series()
    weights = model.data['_weights'].to_series()

    cost_getter = utils.cost_getter(model.get_option)
    depreciation_getter = utils.depreciation_getter(model.get_option)
    cost_per_distance_getter = utils.cost_per_distance_getter(model.config_model)

    @utils.memoize
    def _depreciation_rate(y, k):
        return depreciation_getter(y, k)

    @utils.memoize
    def _cost(cost, y, k, x=None):
        return cost_getter(cost, y, k, x=x)

    @utils.memoize
    def _cost_per_distance(cost, y, k, x):
        return cost_per_distance_getter(cost, y, k, x)

    def _check_and_set(cost, y, x, k):
        """
        Ensure that sufficient constraints have been set to allow negative
        costs, where applicable.
        Returns cost if bounds are set, raises error if unset
        """
        e = exceptions.OptionNotSetError
        if y in m.y_transmission:
            # Divided by 2 for transmission techs because construction costs
            # are counted at both ends
            unit_cost = (_cost(cost, y, k, x) + _cost_per_distance(cost, y, k, x)) / 2
        else:
            unit_cost = _cost(cost, y, k, x)

        # Search the Pyomo constraint for this technology and location, see if it's set

        if (y, x) in getattr(m, 'c_' + cost).keys() or unit_cost >= 0:
            return unit_cost * getattr(m, cost)[y, x]
        elif unit_cost < 0:
            raise e(cost + '.max must be defined for {}:{} '
                    'as cost is negative'.format(y, x))

    def cost_fixed_rule(m, y, x, k):
        if y in m.y_store and x in m.x_store:
            cost_s_cap = _check_and_set('s_cap', y, x, k)
        else:
            cost_s_cap = 0

        if y in m.y_sp_finite_r and x in m.x_r:
            cost_r_cap = _check_and_set('r_cap', y, x, k)
        else:
            cost_r_cap = 0

        if y in m.y_r_area and x in m.x_r:
            cost_r_area = _check_and_set('r_area', y, x, k)
        else:
            cost_r_area = 0

        cost_e_cap = _check_and_set('e_cap', y, x, k)

        if y in m.y_sp_r2 and x in m.x_r:
            cost_r2_cap = _check_and_set('r2_cap', y, x, k)
        else:
            cost_r2_cap = 0

        cost_con = (_depreciation_rate(y, k) *
            (sum(time_res * weights) / 8760) *
            (cost_s_cap + cost_r_cap + cost_r_area + cost_r2_cap +
             cost_e_cap)
        )
        if _cost('om_fixed', y, k, x) < 0 and (y, x) not in m.c_e_cap.keys():
            raise exceptions.OptionNotSetError(
                'e_cap.max must be defined '
                'for {}:{} as `om_fixed` cost is negative'.format(y, x))

        return (m.cost_fixed[y, x, k] ==
                    _cost('om_frac', y, k, x) * cost_con
                    + (_cost('om_fixed', y, k, x) * m.e_cap[y, x] *
                       (sum(time_res * weights) / 8760)) + cost_con)

    def cost_var_rule(m, y, x, t, k):
        om_var = get_cost_param(model, 'om_var', k, y, x, t)
        om_fuel = get_cost_param(model, 'om_fuel', k, y, x, t)
        if y in m.y_export and x in m.x_export:
            export = m.export[y, x, t]
            cost_export = get_cost_param(model, 'export', k, y, x, t) * export
        else:
            export = 0
            cost_export = 0
        carrier = model.get_option(y + '.carrier', default=y + '.carrier_out')
        if isinstance(carrier, dict):  # Conversion_plus with multiple primary output carriers
            carrier = model.get_option(y + '.primary_carrier')
            if not carrier:
                raise exceptions.OptionNotSetError(
                    'No specific carrier set for attributing variable costs '
                    'to conversion+ tech {} at {}'.format(y, x))
        # Note: only counting c_prod for operational costs.
        # This should generally be a reasonable assumption to make.
        # It might be necessary to remove parasitic losses for this
        # i.e. c_prod --> es_prod.
        if om_var:
            cost_op_var = om_var * weights.loc[t] * m.c_prod[carrier, y, x, t]
        else:
            cost_op_var = 0

        cost_op_fuel = 0
        if y in m.y_supply_plus and om_fuel:
            r_eff = get_constraint_param(model, 'r_eff', y, x, t)
            if po.value(r_eff) > 0:  # In case r_eff is zero, to avoid an infinite value
                # Dividing by r_eff here so we get the actual r used, not the r
                # moved into storage...
                cost_op_fuel = (om_fuel * weights.loc[t] * (m.r[y, x, t] / r_eff))
        elif y in m.y_supply and om_fuel:  # m.r == m.c_prod/e_eff
            e_eff = get_constraint_param(model, 'e_eff', y, x, t)
            if po.value(e_eff) > 0:  # in case e_eff is zero, to avoid an infinite value
                cost_op_fuel = (om_fuel * weights.loc[t] *
                                (m.c_prod[carrier, y, x, t] / e_eff))

        cost_op_r2 = 0
        if y in m.y_sp_r2:
            r2_eff = get_constraint_param(model, 'r2_eff', y, x, t)
            if po.value(r2_eff) > 0:  # in case r2_eff is zero, to avoid an infinite value
                om_r2 = get_cost_param(model, 'om_r2', k, y, x, t)
                cost_op_r2 = (om_r2 * weights.loc[t] * (m.r2[y, x, t] / r2_eff))

        return (m.cost_var[y, x, t, k] ==
                cost_op_var + cost_op_fuel + cost_op_r2 + cost_export)

    def cost_rule(m, y, x, k):
        return (
            m.cost[y, x, k] ==
            m.cost_fixed[y, x, k] +
            sum(m.cost_var[y, x, t, k] for t in m.t)
        )

    # Constraints
    m.c_cost_fixed = po.Constraint(m.y, m.x, m.k, rule=cost_fixed_rule)
    m.c_cost_var = po.Constraint(m.y, m.x, m.t, m.k, rule=cost_var_rule)
    m.c_cost = po.Constraint(m.y, m.x, m.k, rule=cost_rule)


def model_constraints(model):
    m = model.m

    @utils.memoize
    def get_parents(level):
        return list(model._locations[model._locations._level == level].index)

    @utils.memoize
    def get_children(parent, childless_only=True):
        """
        If childless_only is True, only children that have no children
        themselves are returned.

        """
        locations = model._locations
        children = list(locations[locations._within == parent].index)
        if childless_only:  # FIXME childless_only param needs tests
            children = [i for i in children if len(get_children(i)) == 0]
        return children

    # Constraint rules
    def c_system_balance_rule(m, c, x, t):
        # Balacing takes place at top-most (level 0) locations, as well
        # as within any lower-level locations that contain children
        if (model._locations.at[x, '_level'] == 0
                or len(get_children(x)) > 0):
            family = get_children(x) + [x]  # list of children + parent
            balance = (sum(m.c_prod[c, y, xs, t]
                           for xs in family for y in m.y)
                       + sum(m.c_con[c, y, xs, t]
                             for xs in family for y in m.y)
                       - sum(m.export[y, xs, t]
                             for xs in family for y in m.y_export if xs in m.x_export
                             and c == model.get_option(y + '.export', x=xs)))

            return balance == 0
        else:
            return po.Constraint.NoConstraint

    # Constraints
    m.c_system_balance = po.Constraint(m.c, m.x, m.t,
                                       rule=c_system_balance_rule)

    def c_export_balance_rule(m, c, y, x, t):
        # Ensuring no technology can 'pass' its export capability to another
        # technology with the same carrier_out,
        # by limiting its export to the capacity of its production
        if c == model.get_option(y + '.export', x=x):
            return m.c_prod[c, y, x, t] >= m.export[y, x, t]
        else:
            return po.Constraint.Skip

    # Constraints
    m.c_export_balance = po.Constraint(m.c, m.y_export, m.x_export, m.t,
                                       rule=c_export_balance_rule)



import pytest
import tempfile

from calliope import examples
from calliope.utils import AttrDict
from . import common
from .common import assert_almost_equal, solver, solver_io

def create_and_run_model(override=""):
    locations = """
        locations:
            1:
                techs: ['ccgt', 'demand_power', 'unmet_demand_power']
                override:
                    ccgt:
                        constraints:
                            e_cap.max: 40
                    demand_power:
                        constraints:
                            r: -10
            2:
                techs: ['ccgt', 'demand_power', 'unmet_demand_power']
                override:
                    ccgt:
                        constraints:
                            e_cap.max: 30
                        costs:
                            e_cap: 3000
                    demand_power:
                        constraints:
                            r: -10
        links:
            1,2:
                hvac:
                    constraints:
                        e_cap.max: 100
    """
    config_run = """
        mode: plan
        model: ['{techs}', '{locations}']
        subset_t: ['2005-01-01', '2005-01-01']
    """
    override = AttrDict.from_yaml_string(override)
    override.set_key('solver', solver)
    override.set_key('solver_io', solver_io)
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(locations.encode('utf-8'))
        f.read()
        model = common.simple_model(config_run=config_run,
                                    config_locations=f.name,
                                    override=override)
    model.run()
    return model

class TestModel:
    @pytest.fixture(scope='module')
    def model(self, override=""):
        locations = """
            locations:
                1:
                    techs: []
                2:
                    techs: ['demand_power']
                    override:
                        demand_power:
                            constraints:
                                r: -90
                sub1,sub2:
                    within: 1
                    techs: ['ccgt']
                    override:
                        ccgt:
                            constraints:
                                e_cap.max: 60
            links:
                1,2:
                    hvac:
                        constraints:
                            e_eff: 0.90
                            e_cap.max: 100
        """
        config_run = """
            mode: plan
            model: ['{techs}', '{locations}']
            subset_t: ['2005-01-01', '2005-01-02']
        """
        override = AttrDict.from_yaml_string(override)
        override.set_key('solver', solver)
        override.set_key('solver_io', solver_io)
        with tempfile.NamedTemporaryFile(delete=False) as f:
            f.write(locations.encode('utf-8'))
            f.read()
            model = common.simple_model(config_run=config_run,
                                        config_locations=f.name,
                                        override=override)
        model.run()
        return model

    def test_model_solves(self, model):
         assert str(model.results.solver.termination_condition) == 'optimal'

    def test_model_balanced(self, model):
        sol = model.solution
        assert sol['e'].loc[dict(c='power', y='ccgt')].sum(dim='x').mean() == 100
        assert (sol['e'].loc[dict(c='power', y='hvac:1')].sum(dim='x') ==
                -1 * sol['e'].loc[dict(c='power', y='demand_power')].sum(dim='x')).all()

    def test_model_costs(self, model):
        sol = model.solution
        assert_almost_equal(sol['summary'].to_pandas().loc['ccgt', 'levelized_cost_monetary'], 0.1)

    def test_one_way(self):
        """
        Check that one_way transmission can be forced using 'one_way' in model
        formulation.
        """
        override="""
            override.links:
                    X1,N1:
                        heat_pipes:
                            constraints:
                                one_way: true
                    N1,X2:
                        heat_pipes:
                            constraints:
                                one_way: true
                    N1,X3:
                        heat_pipes:
                            constraints:
                                one_way: true
        """
        model = examples.UrbanScale(override=AttrDict.from_yaml_string(override))

        model.run()
        sol = model.solution

        # Usual urban scale model has non-zero transmission along each of these
        # links, one_way forces them to zero
        assert_almost_equal(sol.c_prod.loc[
            dict(y='heat_pipes:N1', c='heat', x='X2')], 0, 0.1)
        assert_almost_equal(sol.c_prod.loc[
            dict(y='heat_pipes:N1', c='heat', x='X3')], 0, 0.1)
        assert_almost_equal(sol.c_prod.loc[
            dict(y='heat_pipes:X1', c='heat', x='N1')], 0, 0.1)

        # Usual urban scale model has zero transmission along this link,
        # one_way forces a change in optimal solution, making it non-zero
        assert_almost_equal(sol.c_prod.loc[
            dict(y='heat_pipes:X2', c='heat', x='N1')], 2.6, 0.1)



import pytest
import tempfile

import calliope
from calliope.utils import AttrDict
from . import common
from .common import solver, solver_io

def create_and_run_model(override):
    locations = """
        locations:
            1:
                techs: ['ccgt', 'demand_power',
                        'unmet_demand_power', 'csp']
                override:
                    demand_power:
                        x_map: '1: demand'
                        constraints:
                            r: file=demand-sin_r.csv
        links:
    """
    config_run = """
        mode: plan
        model: ['{techs}', '{locations}']
        subset_t: ['2005-01-01', '2005-01-02']
    """

    override = AttrDict.from_yaml_string(override)
    override.set_key('solver', solver)
    override.set_key('solver_io', solver_io)
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(locations.encode('utf-8'))
        f.read()
        model = common.simple_model(config_run=config_run,
                                    config_locations=f.name,
                                    override=override)
    model.run()
    return model

class TestModel:

    def test_model_export(self):
        override = """
            override.techs.ccgt:
                        export: True
                        costs:
                            monetary:
                                export: -0.2
        """
        model = create_and_run_model(override)
        assert str(model.results.solver.termination_condition) == 'optimal'

    def test_model_fixed_rev(self):
        override = """
            override.techs.ccgt.costs.monetary.e_cap: -5
        """
        model = create_and_run_model(override)
        assert str(model.results.solver.termination_condition) == 'optimal'

    def test_model_var_rev(self):
        override = """
            override.techs.ccgt.costs.monetary.om_var: -0.1
        """
        model = create_and_run_model(override)
        assert str(model.results.solver.termination_condition) == 'optimal'

    def test_model_all_rev(self):
        override = """
            override.techs.ccgt:
                        export: True
                        costs.monetary:
                                om_var: -0.1
                                e_cap: -5
                                export: -0.2
        """
        model = create_and_run_model(override)
        assert str(model.results.solver.termination_condition) == 'optimal'

    def test_model_invalid_r_cap(self):
        override = """
            override.techs.csp.costs.monetary.r_cap: -5
        """
        with pytest.raises(calliope.exceptions.OptionNotSetError):
            model = create_and_run_model(override)

    def test_model_valid_r_cap(self):
        override = """
            override.techs.csp:
                        constraints.r_cap.max: 10
                        costs.monetary.r_cap: -5
        """
        model = create_and_run_model(override)
        assert str(model.results.solver.termination_condition) == 'optimal'



import tempfile

from calliope.utils import AttrDict
from . import common
from .common import solver, solver_io, _add_test_path

def create_and_run_model(override, iterative_warmstart=False):
    locations = """
        locations:
            1:
                techs: ['ccgt', 'demand_power']
                override:
                    ccgt:
                        constraints:
                            e_cap.max: 100
                    demand_power:
                        constraints:
                            r: -10
        links:
    """
    config_run = """
        mode: plan
        model: ['{techs}', '{locations}']
        subset_t: ['2005-01-01', '2005-01-02']
    """
    override = AttrDict.from_yaml_string(override)
    override.set_key('solver', solver)
    override.set_key('solver_io', solver_io)
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(locations.encode('utf-8'))
        f.read()
        model = common.simple_model(config_run=config_run,
                                    config_locations=f.name,
                                    override=override,
                                    path=_add_test_path('common/t_constraints_from_file'))
    model.run(iterative_warmstart)
    return model

class TestModel:

    # all constraints are fixed values
    def test_model_fixed(self):
        override = """override:"""
        model = create_and_run_model(override)
        assert str(model.results.solver.termination_condition) == 'optimal'

    # e_eff is a timeseries variable
    def test_model_var_e_eff(self):
        override = """
            override:
                techs:
                    ccgt:
                        constraints:
                            e_eff: file=eff_var_sin.csv
                    """
        model = create_and_run_model(override)
        assert str(model.results.solver.termination_condition) == 'optimal'

    # r_eff is a timeseries variable
    def test_model_var_r_eff(self):
        override = """
            override:
                techs:
                    ccgt:
                        constraints:
                            r_eff: file=eff_var_sin.csv
                    """
        model = create_and_run_model(override)
        assert str(model.results.solver.termination_condition) == 'optimal'

    # costs are timeseries variables
    def test_model_var_om_var(self):
        override = """
            override:
                techs:
                    ccgt:
                        costs:
                            monetary:
                                om_fuel: file=cost_var.csv
                                om_var: file=cost_var.csv
                    """
        model = create_and_run_model(override)
        assert str(model.results.solver.termination_condition) == 'optimal'


    # revenue is a timeseries variable
    def test_model_var_sub_var(self):
        override = """
            override.techs.ccgt.costs.monetary.om_var: file=rev_var.csv
                    """
        model = create_and_run_model(override)
        assert str(model.results.solver.termination_condition) == 'optimal'



import calliope


class TestNationalScaleExampleModel:
    def test_model_initialization_default(self):
        model = calliope.examples.NationalScale()
        model.run()

class TestUrbanScaleExampleModel:
    def test_model_initialization_default(self):
        model = calliope.examples.UrbanScale()
        model.run()



import pytest
import tempfile

from calliope.utils import AttrDict
from . import common
from .common import solver, solver_io, assert_almost_equal

def create_and_run_model(override=""):
    locations = """
        locations:
            1:
                techs: ['ccgt', 'test_conversion', 'test_conversion_plus',
                        'demand_power', 'unmet_demand_power',
                        'demand_heat', 'unmet_demand_heat',
                        'demand_low_T', 'unmet_demand_low_T',
                        'demand_V_low_T', 'unmet_demand_V_low_T']
                override:
                    demand_power:
                            x_map: '1: demand'
                            constraints:
                                r_scale_to_peak: -10
                                r: file=demand-sin_r.csv
                    demand_heat:
                            x_map: '1: demand'
                            constraints:
                                r_scale_to_peak: -5
                                r: file=demand-sin_r.csv
                    demand_low_T:
                            constraints:
                                r: -12
                    demand_V_low_T:
                            constraints:
                                r: -12
        links:
    """
    config_run = """
        mode: plan
        model: ['{techs}', '{locations}']
        subset_t: ['2005-01-01', '2005-01-01']
    """

    override = AttrDict.from_yaml_string(override)
    override.set_key('solver', solver)
    override.set_key('solver_io', solver_io)
    with tempfile.NamedTemporaryFile(delete=False) as f:
        f.write(locations.encode('utf-8'))
        f.read()
        model = common.simple_model(config_run=config_run,
                                    config_locations=f.name,
                                    override=override)
    model.run()
    return model

class TestModel:
    def test_export_from_supply(self):
        # giving export revenue *should* lead to there being an export of power
        # and a higher consumption of resource for the ccgt
        override = """
                    override.techs.ccgt:
                                export: true
                                costs.monetary.export: -10
                    """
        model1 = create_and_run_model()
        model2 = create_and_run_model(override)
        sol1 = model1.solution
        sol2 = model2.solution
        with pytest.raises(AttributeError):
            sol1.export # there should be no export here
        assert_almost_equal(sol1.c_con.sum(), sol2.c_con.sum())
        assert_almost_equal(sol2.export.sum(), 687.5, tolerance=0.01)

    def test_export_from_conversion(self):
        # conversion technology exporting heat
        override = """
                    override.techs.test_conversion:
                                export: true
                                costs.monetary.export: -10
                    """
        model = create_and_run_model(override)
        assert_almost_equal(model.solution.export.sum(), 1375, tolerance=0.01)

    def test_export_from_conversion_plus(self):
        # conversion_plus technology exporting low_T (primary carrier), V_low_T, and power
        override1 = """
                    override.techs.test_conversion_plus:
                                export: true # will default to primary carrier
                                costs.monetary.export: -10
                    """
        override2 = """
                    override.techs.test_conversion_plus:
                                export: V_low_T
                                costs.monetary.export: -10
                    """
        override3 = """
                    override.techs.test_conversion_plus:
                                export: power
                                costs.monetary.export: -100
                    """
        model1 = create_and_run_model(override1)
        assert_almost_equal(model1.solution.export.sum(), 1901.6, tolerance=0.01)
        assert_almost_equal(model1.solution.c_con.loc[dict(y='demand_low_T', c='low_T')].sum()
                             - model1.solution.export.sum(),
                             -model1.solution.c_prod.loc[dict(y='test_conversion_plus', c='low_T')].sum())

        model2 = create_and_run_model(override2)
        assert_almost_equal(model2.solution.export.sum(), 1521.28, tolerance=0.01)
        assert_almost_equal(model2.solution.c_con.loc[dict(c='V_low_T')].sum()
                             - model2.solution.export.sum(),
                             -model2.solution.c_prod.loc[dict(y='test_conversion_plus', c='V_low_T')].sum())

        model3 = create_and_run_model(override3)
        assert_almost_equal(model3.solution.export.sum(), 32.4, tolerance=0.01)
        assert_almost_equal(model3.solution.c_con.loc[dict(c='power')].sum()
                             - model3.solution.export.sum(),
                             -model3.solution.c_prod.loc[dict(c='power')].sum())



"""
Copyright (C) 2013-2017 Calliope contributors listed in AUTHORS.
Licensed under the Apache 2.0 License (see LICENSE file).

sets.py
~~~~~~~

Sets & sub-sets defined to reduce size of decision variables & constraints.

The markers preceded with `###` are used to auto-include the set description
in the documentation

Sets are:

###PART TO INCLUDE IN DOCUMENTATION STARTS HERE###

Main sets & sub-sets
====================

Technologies:

* ``m.y_demand``: all demand sources
    * ``m.y_sd_r_area``: if any r_area constraints are defined (shared)
    * ``m.y_sd_finite_r``: if finite resource limit is defined (shared)
* ``m.y_supply``: all basic supply technologies
    * ``m.y_sd_r_area``: if any r_area constraints are defined (shared)
    * ``m.y_sd_finite_r``: if finite resource limit is defined (shared)
* ``m.y_storage``: specifically storage technologies
* ``m.y_supply_plus``: all supply+ technologies
    * ``m.y_sp_r_area``: If any r_area constraints are defined
    * ``m.y_sp_finite_r``: if finite resource limit is defined
    * ``m.y_sp_r2``: if secondary resource is allowed
* ``m.y_conversion``: all basic conversion technologies
* ``m.y_conversion_plus``: all conversion+ technologies
    * ``m.y_cp_2out``: secondary carrier(s) out
    * ``m.y_cp_3out``: tertiary carrier(s) out
    * ``m.y_cp_2in``: secondary carrier(s) in
    * ``m.y_cp_3in``: tertiary carrier(s) in
* ``m.y_transmission``: all transmission technologies
* ``m.y_unmet``: dummy supply technologies to log

Locations:

* ``m.x_trans``: all transmission locations
* ``m.x_r``: all locations which act as system sources/sinks
* ``m.x_store``: all locations in which storage is allowed

Shared subsets
==============

* ``m.y_finite_r``: shared between y_demand, y_supply, and y_supply_plus. Contains:
    * ``m.y_sd_finite_r``
    * ``m.y_sp_finite_r``
* ``m.y_r_area``: shared between y_demand, y_supply, and y_supply_plus. Contains:
    * ``m.y_sd_r_area``
    * ``m.y_sp_r_area``

Meta-sets
=========

Technologies:

* ``m.y``: all technologies, includes:
    * ``m.y_demand``
    * ``m.y_supply``
    * ``m.y_storage``
    * ``m.y_supply_plus``
    * ``m.y_conversion``
    * ``m.y_conversion_plus``
    * ``m.y_transmission``
* ``m.y_sd``: all basic supply & demand technologies, includes:
    * ``m.y_demand``
    * ``m.y_supply``
* ``m.y_store``: all technologies that have storage capabilities, includes:
    * ``m.y_storage``
    * ``m.y_supply_plus``

Locations:

* ``m.x``: all locations, includes:
    * ``m.x_trans``
    * ``m.x_r``
    * ``m.x_store``

###PART TO INCLUDE IN DOCUMENTATION ENDS HERE###

"""

import numpy as np

from . import exceptions
from . import transmission


def init_set_y(model, _x):
    """
    Initialise sets and subsets of y
    """

    # Subset of transmission technologies, if any defined
    # Used to initialize transmission techs further below
    # (not yet added to main `y` set here)
    links = model.config_model.get_key('links', None)
    if links:
        _y_trans = transmission.get_transmission_techs(links)
        transmission_techs = list(set([list(v.keys())[0]
                                  for k, v in links.items()]))
        _x_trans = set()
        for link in links.keys():
            locs = link.split(',')
            _x_trans.add(locs[0])
            _x_trans.add(locs[1])
        _x_trans = list(_x_trans)
    else:
        _y_trans = []
        _x_trans = []
        transmission_techs = []

    _y = set()
    _x = _x.copy()
    for k, v in model.config_model.locations.items():
        if 'techs' in v.keys():
            if set(v.techs).intersection(transmission_techs): # transmission nodes have only tranmission techs in techs list
                _x.remove(k)
            else:
                for y in v.techs:
                    if y in model.config_model.techs:
                        _y.add(y)
                    else:
                        e = exceptions.ModelError
                        raise e('Location `{}` '
                                'uses undefined tech `{}`.'.format(k, y))
        else:
            e = exceptions.ModelError
            raise e('The region `' + k + '` does not allow any '
                    'technologies via `techs` nor does it have '
                    'links to other locations. Must give at '
                    'least one technology per region or link '
                    'to other regions.')
    _y = list(_y)

    # Potentially subset _y
    if model.config_run.get_key('subset_y', default=False):
        _y = [y for y in _y if y in model.config_run.subset_y]

    # Subset of technologies that are demand sources
    _y_demand = [y for y in _y if model.ischild(y, of='demand')]

    # Subset of supply technologies
    _y_supply = [y for y in _y if model.ischild(y, of='supply')]

    # Subset of supply_plus technologies
    _y_supply_plus = [y for y in _y if model.ischild(y, of='supply_plus')]

    # Subset of storage technologies
    _y_storage = [y for y in _y if model.ischild(y, of='storage')]

    # Subset of conversion technologies

    _y_conversion = [y for y in _y if model.ischild(y, of='conversion')]

    # Subset of conversion+ technologies
    _y_conversion_plus = [y for y in _y if model.ischild(y, of='conversion_plus')]

    # Subset of unmet technologies
    _y_unmet = [y for y in _y if model.ischild(y, of='unmet_demand')]

    # Subset of basic supply & demand technologies
    _y_sd = _y_supply + _y_demand + _y_unmet

    # Subset of storage technologies
    _y_store = [y for y in np.concatenate((_y_storage,_y_supply_plus)) if
                any([model.get_option(y + '.constraints.s_cap.max', x=x) +
                     model.get_option(y + '.constraints.s_time.max', x=x) +
                     model.get_option(y + '.constraints.s_cap.equals', x=x) +
                     model.get_option(y + '.constraints.c_rate', x=x)
                     for x in _x])]

    _y_sd_r_area = [y for y in _y_sd if
                    any([model.get_option(y + '.constraints.r_area.max', x=x) +
                         model.get_option(y + '.constraints.r_area.equals', x=x) +
                         model.get_option(y + '.constraints.r_area_per_e_cap', x=x)
                         for x in _x])]
    _y_sp_r_area = [y for y in _y_supply_plus if
                    any([model.get_option(y + '.constraints.r_area.max', x=x) +
                         model.get_option(y + '.constraints.r_area.equals', x=x) +
                         model.get_option(y + '.constraints.r_area_per_e_cap', x=x)
                         for x in _x])]

    _y_r_area = _y_sd_r_area + _y_sp_r_area

    _y_sp_r2 = [y for y in _y_supply_plus if
               any([model.get_option(y + '.constraints.allow_r2', x=x)
                    for x in _x])]

    _y_cp_2out = [y for y in _y_conversion_plus
                  if model.get_option(y + '.carrier_out_2')]
    _y_cp_3out = [y for y in _y_conversion_plus
                  if model.get_option(y + '.carrier_out_3')]
    _y_cp_2in = [y for y in _y_conversion_plus
                 if model.get_option(y + '.carrier_in_2')]
    _y_cp_3in = [y for y in _y_conversion_plus
                 if model.get_option(y + '.carrier_in_3')]

    # subset of technologies allowing export
    _y_export = [y for y in _y
                 if any([model.get_option(y + '.export', x=x) for x in _x])]
    sets = {
        'y': _y,
        'y_demand': _y_demand,
        'y_supply': _y_supply,
        'y_supply_plus': _y_supply_plus,
        'y_storage': _y_storage,
        'y_conversion': _y_conversion,
        'y_conversion_plus': _y_conversion_plus,
        'y_unmet': _y_unmet,
        'y_sd': _y_sd,
        'y_store': _y_store,
        'y_sd_r_area': _y_sd_r_area,
        'y_sp_r_area': _y_sp_r_area,
        'y_r_area': _y_r_area,
        'y_sp_r2': _y_sp_r2,
        'y_cp_2out': _y_cp_2out,
        'y_cp_3out': _y_cp_3out,
        'y_cp_2in': _y_cp_2in,
        'y_cp_3in': _y_cp_3in,
        'y_transmission': _y_trans,
        'techs_transmission': transmission_techs,
        'x_transmission': _x_trans,
        'y_export': _y_export,
        'x_transmission_plus': _x
    }


    return sets

def init_set_x(model):
    """
    Add subsets of x. x_transmission already set in init_set_y
    This function has to be run *after* init_set_y to ensure technology sets are
    already included in model._sets.
    """
    locations = model.config_model.locations
    # All locations which interact with the spatial boundary
    _x_r = [x for x in model._sets['x'] if
            set(model._sets['y_supply_plus'] +
            model._sets['y_sd']).intersection(locations[x].techs)]
    # All locations in which energy is stored
    _x_store = [x for x in model._sets['x'] if
                set(model._sets['y_store']).intersection(locations[x].techs)]
    # All locations which demand energy
    _x_demand = [x for x in model._sets['x'] if
                 set(model._sets['y_demand']).intersection(locations[x].techs)]
    # All locations in which energy is converted between carriers
    _x_conversion = [x for x in model._sets['x'] if
                     set(model._sets['y_conversion_plus'] +
                     model._sets['y_conversion']).intersection(locations[x].techs)]
    # All locations in which energy is exported
    _x_export = [x for x in model._sets['x'] if
                     set(model._sets['y_export']).intersection(locations[x].techs)]

    return {
        'x_r': _x_r,
        'x_store': _x_store,
        'x_demand': _x_demand,
        'x_conversion': _x_conversion,
        'x_export': _x_export
    }

def init_set_c(model):
    """
    Add all energy carriers to the set `c`.
    """
    def _check_if_dict(option):
        if isinstance(option, dict):
            return option.keys()
        else:
            return [option]
    _c = set()
    sets = model._sets
    for y in sets['y']:  # Only add carriers for allowed technologies
        if y in sets['y_supply'] or y in sets['y_supply_plus']:
            _c.update([model.get_option(y + '.carrier',
                                default=y + '.carrier_out')])
        if y in sets['y_demand']:
            _c.update([model.get_option(y + '.carrier',
                                default=y + '.carrier_in')])
        if y in sets['y_transmission'] or y in sets['y_storage']:
            _c.update([model.get_option(y + '.carrier')])
        if y in sets['y_conversion']:
            _c.update([model.get_option(y + '.carrier_in')])
            _c.update([model.get_option(y + '.carrier_out')])
        if y in sets['y_conversion_plus']:
            _c.update(_check_if_dict(model.get_option(y + '.carrier_in')))
            _c.update(_check_if_dict(model.get_option(y + '.carrier_out')))
        if y in sets['y_cp_2out']:
            _c.update(_check_if_dict(model.get_option(y + '.carrier_out_2')))
        if y in sets['y_cp_3out']:
            _c.update(_check_if_dict(model.get_option(y + '.carrier_out_3')))
        if y in sets['y_cp_2in']:
            _c.update(_check_if_dict(model.get_option(y + '.carrier_in_2')))
        if y in sets['y_cp_3in']:
            _c.update(_check_if_dict(model.get_option(y + '.carrier_in_3')))
    _c = _c.difference([True, False])
    return list(_c)



def init_y_trans(model):
    # Add transmission technologies to y, if any defined
    _y_trans = model._sets['y_transmission']

    if _y_trans:
        model._sets['y'].extend(_y_trans)

        # Add transmission tech columns to locations matrix
        for y in _y_trans:
            model._locations[y] = 0

        # Create representation of location-tech links
        tree = transmission.explode_transmission_tree(
            model.config_model.links, model._sets['x_transmission']
        )

        # Populate locations matrix with allowed techs and overrides
        if tree:
            for x in tree:
                for y in tree[x]:
                    # Allow the tech
                    model._locations.at[x, y] = 1
                    # Add constraints if needed. Speficying one way transmission
                    # leads to e_prod in the direction opposite to the defined
                    # link being set to False. E.g. if 'a,b' is one way for tech
                    # 'y' then y:a.constraints.e_prod @ x=b -> False
                    for c in tree[x][y].keys_nested():
                        if c == 'constraints.one_way':
                            colname = '_override.' + y + '.constraints.e_prod'
                        else:
                            colname = '_override.' + y + '.' + c
                        if colname not in model._locations.columns:
                            model._locations[colname] = np.nan
                        if (c == 'constraints.one_way'
                                and '{},{}'.format(y.split(':')[1], x) in
                                model.config_model.links.keys()):
                            # set the option in model._locations and model.config_model.locations
                            model.set_option(y + '.constraints.e_prod', False, x=x)
                        else:
                            model._locations.at[x, colname] = tree[x][y].get_key(c)
